/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.framework.server;

import java.io.Serializable;

import java.util.Map;
import java.util.Iterator;

import org.apache.commons.lang.builder.ToStringBuilder;


/**
 * This class represents a Sync4j device.
 *
 * @author  Luigia Fassina @ Funambol
 *
 * @version $Id: Sync4jDevice.java,v 1.11 2005/06/13 17:16:59 nichele Exp $
 *
 */
public class Sync4jDevice implements Serializable {

    // --------------------------------------------------------------- Constants
    private static final String DEFAULT_CHARSET = "UTF-8";

    // ------------------------------------------------------------ Private data
    private String deviceId          ;
    private String description       ;
    private String type              ;
    private byte[] clientNonce       ;
    private byte[] serverNonce       ;
    private String serverPassword    ;
    private String timeZone          ;
    private boolean convertDate      ;
    private Capabilities capabilities;
    private String charset           ;


    // ------------------------------------------------------------- Costructors
    /** Creates a new instance of Sync4jDevice */
    public Sync4jDevice() {
        this(null, null, null, null, null, null, null, false, null);
    }

    /**
     * Creates a new instance of Sync4jDevice.
     *
     * @param deviceId the device identification
     */
    public Sync4jDevice(final String deviceId) {
        this(deviceId, null, null, null, null, null, null, false, null);
    }

    /**
     * Creates a new instance of Sync4jDevice.
     *
     * @param deviceId the device identification
     * @param description the device description
     * @param type the device type
     *
     */
    public Sync4jDevice(final String deviceId   ,
                        final String description,
                        final String type       ) {
        this(deviceId, description, type, null, null, null, null, false, null);

    }

    /**
     * Creates a new instance of Sync4jDevice.
     *
     * @param deviceId the device identification
     * @param description the device description
     * @param type the device type
     * @param clientNonce the next nonce sent by server for MD5 authentication
     * @param serverNonce the next nonce sent by client for MD5 authentication
     */
    public Sync4jDevice(final String deviceId   ,
                        final String description,
                        final String type       ,
                        final byte[] clientNonce ,
                        final byte[] serverNonce) {
        this(deviceId,
             description,
             type,
             clientNonce,
             serverNonce,
             null,
             null,
             false,
             null);

    }

    /**
     * Creates a new instance of Sync4jDevice.
     *
     * @param deviceId the device identification
     * @param description the device description
     * @param type the device type
     * @param clientNonce    the next nonce sent by server for MD5 authentication
     * @param serverNonce    the next nonce sent by client for MD5 authentication
     * @param serverPassword the server password for this device
     * @param timeZone       the timezone for this device
     * @param convertDate    must the date be converted in the device timezone ?
     */
    public Sync4jDevice(final String  deviceId   ,
                        final String  description,
                        final String  type       ,
                        final byte[]  clientNonce,
                        final byte[]  serverNonce,
                        final String  serverPassword,
                        final String  timeZone,
                        final boolean convertDate,
                        final String  charset) {
        this.deviceId       = deviceId;
        this.description    = description;
        this.type           = type;
        this.clientNonce    = clientNonce;
        this.serverNonce    = serverNonce;
        this.serverPassword = serverPassword;
        this.timeZone       = timeZone;
        this.convertDate    = convertDate;
        if (charset == null) {
            this.charset    = DEFAULT_CHARSET;
        } else {
            this.charset    = charset;
        }
        
        //
        // Default capabilities...
        //
        this.capabilities = new Capabilities();
    }


    /** Getter for property deviceId.
     * @return Value of property deviceId.
     *
     */
    public String getDeviceId() {
        return deviceId;
    }

    /** Setter for property deviceId.
     * @param deviceId New value of property deviceId.
     *
     */
    public void setDeviceId(String deviceId) {
        this.deviceId = deviceId;
    }

    /** Getter for property description.
     * @return Value of property description.
     *
     */
    public String getDescription() {
        return description;
    }

    /** Setter for property description.
     * @param description New value of property description.
     *
     */
    public void setDescription(String description) {
        this.description = description;
    }

    /** Getter for property type.
     * @return Value of property type.
     *
     */
    public String getType() {
        return type;
    }

    /** Setter for property type.
     * @param type New value of property type.
     *
     */
    public void setType(String type) {
        this.type = type;
    }

    /**
     * Gets the client_nonce property
     *
     * @return clientNonce the nonce that the server uses in order to authenticate
     *                     iteself with the client
     */
    public byte[] getClientNonce() {
        return this.clientNonce;
    }

    /**
     * Sets the client_nonce property
     *
     * @param clientNonce the nonce that the server uses in order to authenticate
     *                    iteself with the client
     */
    public void setClientNonce(byte[] clientNonce) {
        if (clientNonce == null) {
            clientNonce = new byte[0];
        }
        this.clientNonce = clientNonce;
    }

    /**
     * Gets the server_nonce property
     *
     * @return serverNonce the nonce that the client uses in order to authenticate
     *                     iteself with the server
     */
    public byte[] getServerNonce() {
        return this.serverNonce;
    }

    /**
     * Sets the server_nonce property
     *
     * @param serverNonce the nonce that the client uses in order to authenticate
     *                    iteself with the server
     */
    public void setServerNonce(byte[] serverNonce) {
        if (serverNonce == null) {
            serverNonce = new byte[0];
        }
        this.serverNonce = serverNonce;
    }

    /**
     * Gets the server password property
     *
     * @return serverPassword the server password for this device
     */
    public String getServerPassword() {
        return this.serverPassword;
    }

    /**
     * Sets the server password property
     *
     * @param serverPassword the server password for this device
     */
    public void setServerPassword(String serverPassword) {
        this.serverPassword = serverPassword;
    }

    /**
     * Sets the client device capabilities
     *
     * @param capabilities the client device capabilities
     */
    public void setCapabilities(Capabilities capabilities) {
        this.capabilities = capabilities;
    }

    /**
     * Gets the client device capabilities
     *
     * @return the client device capabilities
     */
    public Capabilities getCapabilities() {
        return capabilities;
    }

    /**
     * Sets the client device timeZone
     *
     * @param tz the client device timeZone
     */
    public void setTimeZone(String tz) {
        this.timeZone = tz;
    }

    /**
     * Gets the client device timeZone
     *
     * @return the client device timeZone
     */
    public String getTimeZone() {
        return timeZone;
    }

    /**
     * Sets the convertDate
     *
     * @param convertDate the convert date
     */
    public void setConvertDate(boolean convertDate) {
        this.convertDate = convertDate;
    }

    /**
     * Gets the convertDate
     *
     * @return the convertDate
     */
    public boolean getConvertDate() {
        return convertDate;
    }

    /**
     * Sets the charset
     *
     * @param charset the charset
     */
    public void setCharset(String charset) {
        this.charset = charset;
    }

    /**
     * Gets the charset
     *
     * @return the charset
     */
    public String getCharset() {
        return charset;
    }


    public String toString() {
        ToStringBuilder sb = new ToStringBuilder(this);

         sb.append("deviceId",         deviceId);
         sb.append("description",      description);
         sb.append("type",             type);
         if (clientNonce != null) {
             sb.append("client_nonce", new String(clientNonce));
         }
         if (serverNonce != null) {
             sb.append("server_nonce", new String(serverNonce));
         }
         sb.append("serverPassword",   serverPassword);
         sb.append("timeZone",         timeZone);
         sb.append("convertDate",      convertDate);         
         sb.append("charset",          charset);

         return sb.toString();
    }

}
