/**
 * Copyright (C) 2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

package sync4j.server.config;

import java.io.File;
import java.util.HashMap;
import java.util.Map;
import sync4j.framework.config.ConfigurationException;
import sync4j.framework.tools.IOTools;
import sync4j.framework.tools.beans.BeanFactory;

/**
 * This class implements a cache of configuration objects. Objects are stored in
 * the cache associated with their last modification timestamp. A new instance
 * of the object will be returned only if its server bean is newer  of the one 
 * loaded.
 *
 * It also provides cached access to singleton instances. In case of singleton
 * istances, a new instance of a server bean is returned only if its 
 * configuration has changed; otherwise the last created instance is returned.
 *
 * NOTE: the methods getBeanInstance() and getNewBeanInstance() are not
 *       synchronized by design. We are introducing this class to improve 
 *       performance and synchronized methods on frequently used classes
 *       reduce concurrency significantly. From a functionality point of view,
 *       a race condition won't be very negative: only on high load, some 
 *       threads for a very little time would still see the old server bean
 *       configuration even if the configuration file has already changed.
 *       This is acceptable considering the gain in performance.
 *
 * @author Stefano Fornari, Funambol
 */
public class ConfigCache {

    // ------------------------------------------------------------ Private data
    
    /**
     * The hash map containing the cached server beans. Map keys will be the 
     * server bean names; the entry will be the corresponding ServerBeanEntry
     * objects.
     */
    private Map cache;
    
    // ------------------------------------------------------------ Constructors
    
    /**
     * Creates a new ConfigCache
     */
    public ConfigCache() {
        cache = new HashMap();
    }
    
    // ---------------------------------------------------------- Public methods
    
    /**
     * Returns a new instance of the server bean identified by the given name.
     *
     * @param configLoader class loader to be used to load classes and 
     *        configuration objects
     * @param name the server bean name
     * 
     * @return the server bean instance (a new instance is created each time)
     *
     * @throws Configuration exception in case of errors
     *
     */
    public synchronized Object getNewBeanInstance(ClassLoader configLoader, String name)
    throws ConfigurationException {
        ServerBeanEntry entry = (ServerBeanEntry)cache.get(name);
        
        try {
            //
            // We first check if the bean is in the configuration path with
            // the given name. If so, the file is the configuration to use, 
            // but we need to read it only if recently modified.
            // If the file is not found, the bean represents a class or 
            // something else in the classpath that we do not have to deal with.
            //
            String config = null;
            File configFile = new File(Configuration.getConfigPath(), name);
            
            if (!configFile.exists()) {
                return BeanFactory.getBeanInstance(configLoader, name);
            }
            
            if ((entry == null) || (entry.ts < configFile.lastModified())) {
                entry = new ServerBeanEntry(configFile);
                cache.put(name, entry);
            }
            
            return BeanFactory.getBeanInstanceFromConfig(configLoader, entry.config);
        } catch (Exception e) {
            throw new ConfigurationException( "Error in creating an instance of "
                                            + name
                                            , e                                );
        }
    } 
    
    /**
     * Returns an instance of the server bean identified by the given name. Once
     * an instance of the object is created, it is stored in the cache for 
     * future access. A new instance of the bean is created only when its config
     * file has changed.
     *
     * @param configLoader class loader to be used to load classes and 
     *        configuration objects
     * @param name the server bean name
     * 
     * @return the server bean instance (a new instance is created each time)
     *
     * @throws Configuration exception in case of errors
     *
     */
    public synchronized Object getBeanInstance(ClassLoader configLoader, String name)
    throws ConfigurationException {
        ServerBeanEntry entry = (ServerBeanEntry)cache.get(name);
        
        try {
            //
            // We first check if the bean is in the configuration path with
            // the given name. If so, the file is the configuration to use, 
            // but we need to read it only if recently modified.
            // If the file is not found, the bean represents a class or 
            // something else in the classpath that we do not have to deal with.
            //
            String config = null;
            File configFile = new File(Configuration.getConfigPath(), name);
            
            if (!configFile.exists()) {
                return BeanFactory.getBeanInstance(configLoader, name);
            }
            
            if ((entry == null) || (entry.ts < configFile.lastModified())) {
                entry = new ServerBeanEntry(configFile);
                entry.instance = 
                    BeanFactory.getBeanInstanceFromConfig(configLoader, entry.config);
                cache.put(name, entry);
            }
            
            return entry.instance;
        } catch (Exception e) {
            throw new ConfigurationException( "Error in creating an instance of "
                                            + name
                                            , e                                );
        }
    } 
    
    /**
     * Invalidates the entire cache
     */
    public void invalidate() {
        cache.clear();
    }
    
    /**
     * Invalidates the given entry
     *
     * @param name the bean name to invalidate
     */
    public void invalidate(String name) {
        cache.remove(name);
    }
    
    // --------------------------------------------------------- Private methods
    
    // --------------------------------------------------------- Private classes
    
    /**
     * This represents an entry in the cache. It associates a server bean name
     * to:
     * <ul>
     * <li>ts -> the timestamp of when the configuration file was last read</li>
     * <li>config -> the configuration file content
     * <li>instance -> the singleton instance associated to this server bean
     * 
     */
    class ServerBeanEntry {
        
        /**
         * file timestamp
         */
        public long ts;
        public String config;
        public Object instance;
        
        /**
         * Creates a new ServerBeanEntry from the given file
         *
         * @param file the bean file - NOT NULL
         */
        public ServerBeanEntry(File file) {
            ts = file.lastModified();
            try {
                config = IOTools.readFileString(file);
            } catch (Exception e) {
                //
                // just an easy way to let the caller know, even if it is not
                // very elegant.... :)
                //
                config = e.getMessage();
            }
        }
    }
}
