/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.server.session;

import java.util.logging.Logger;
import java.util.logging.Level;

import java.util.*;

import sync4j.framework.config.ConfigurationConstants;
import sync4j.framework.core.*;
import sync4j.framework.database.Database;
import sync4j.framework.engine.SyncItemImpl;
import sync4j.framework.engine.SyncOperation;
import sync4j.framework.engine.SyncEngine;
import sync4j.framework.engine.SyncItemState;
import sync4j.framework.engine.MessageSizeCalculator;
import sync4j.framework.engine.source.MemorySyncSource;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.engine.pipeline.MessageProcessingContext;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.protocol.*;
import sync4j.framework.security.*;
import sync4j.framework.server.Capabilities;
import sync4j.framework.server.ClientMapping;
import sync4j.framework.server.LastTimestamp;
import sync4j.framework.server.SyncTimestamp;
import sync4j.framework.server.error.ServerException;
import sync4j.framework.server.error.MappingException;
import sync4j.framework.server.error.ServerFailureException;
import sync4j.framework.server.error.InvalidCredentialsException;
import sync4j.framework.server.session.SessionHandler;
import sync4j.framework.server.session.SyncState;
import sync4j.framework.server.store.NotFoundException;
import sync4j.framework.server.store.PersistentStore;
import sync4j.framework.server.Sync4jDevice;
import sync4j.framework.tools.Base64;

import sync4j.server.config.Configuration;
import sync4j.server.engine.Sync4jEngine;

/**
 * This class represents the handler for a SyncML session. It coordinates and
 * handles the packages and messages as dictated by the protocol.
 * <p>
 * The entry point is <i>processMessage()</i>, which determines which message is
 * expected and what processing has to be done (depending on the value of
 * <i>currentState</i>). If an error accours, the session goes to the state
 * <i>STATE_ERROR</i>; in this state no other messages but initialization can be
 * performed.
 * <p>
 * In the current implementation separate initialization is required.
 * <p>
 * <i>SessionHandler</i> makes use of a <i>SyncEngine</i> for all
 * tasks not related to the handling of the protocol.
 * See <i>sync4j.framework.engine.SyncEngine</i> for more information.
 *
 * LOG NAME: sync4j.handler
 *
 * @see sync4j.framework.engine.SyncEngine
 *
 * @author Stefano Fornari @ Funambol
 *
 * @version $Id: SyncSessionHandler.java,v 1.33 2005/10/19 12:03:48 harrie Exp $
 *
 */
public class SyncSessionHandler
implements SessionHandler,
           java.io.Serializable,
           SecurityConstants,
           ConfigurationConstants {

    // ------------------------------------------------------------ Private data

    private int currentState = STATE_START;

    //
    // This data is true for slow sync
    //
    private boolean slow = false;

    /**
     * Gets the current state
     *
     * @return the current state
     */
    public int getCurrentState() {
        return currentState;
    }

    private long creationTimestamp = -1;

    /**
     * Gets the creation timestamp of the session
     *
     * @return the creation timestamp
     */
    public long getCreationTimestamp() {
        return creationTimestamp;
    }

    private transient Logger log = Sync4jLogger.getLogger("handler");

    /**
     * SyncTimestamp for the current synchronization
     */
    private SyncTimestamp nextTimestamp = null;

    private transient SyncInitialization syncInit = null;
    private transient SyncModifications modifications = null;

    /**
     * Contain the client device Id for the current session
     */
    private String clientDeviceId = null;

    /**
     * The databases that have to be synchronized. It is set in the initialization
     * process.
     */
    private Database[] dbs = null;

    /**
     * The type of authetication does impose from the server
     */
    private String clientAuth = null;

    /**
     * Size of the current response message
     */
    private long messageSize = 0;

    /**
     * Which calculator should be used to calculate message size?
     */
    private MessageSizeCalculator sizeCalculator = null;

    // -------------------------------------------------------------- Properties

    /**
     * The session id - read only
     */
    private String sessionId = null;

    public String getSessionId() {
        return this.sessionId;
    }

    /**
     * The command id generator (it defaults ro a <i>CommandIdGenerator</i> instance)
     */
    private CommandIdGenerator cmdIdGenerator = new CommandIdGenerator();

    public void setCommandIdGenerator(CommandIdGenerator cmdIdGenerator) {
        this.cmdIdGenerator = cmdIdGenerator;
    }

    public CommandIdGenerator getCommandIdGenerator() {
        return this.cmdIdGenerator;
    }

    /**
     * The cmdIdGenerator must be reset each time the process
     * of a message is starting
     */
    private void resetIdGenerator() {
        cmdIdGenerator.reset();
        syncEngine.setCommandIdGenerator(cmdIdGenerator);
    }

    /**
     * The message id generator (it defaults ro a <i>SimpleIdGenerator</i> instance)
     */
    private SimpleIdGenerator msgIdGenerator = new SimpleIdGenerator();

    /**
     * The Last message Id from the client
     */
    private String lastMsgIdFromClient = null;

    /**
     * The <i>SyncEngine</i>
     */
    private Sync4jEngine syncEngine = null;

    /**
     * The SyncState for the syncronization process
     */
    private SyncState syncState = null;


    // ---------------------------------------------------------- Public methods
    public SyncEngine getSyncEngine() {
        return this.syncEngine;
    }

    /**
     * Indicates if the session is a new session
     */
    private boolean newSession = true;

    public void setNew(boolean newSession) {
        this.newSession = newSession;
    }

    public boolean isNew() {
        return this.newSession;
    }

    /**
     * Sets the message size calculator
     *
     * @param calculator the message size calculator
     */
    public void setSizeCalculator(MessageSizeCalculator calculator) {
        this.sizeCalculator = calculator;
    }

    // ------------------------------------------------------------ Constructors

    /**
     * Creates a new instance of SyncSessionHandler
     */
    public SyncSessionHandler() {
        this.creationTimestamp = System.currentTimeMillis();
        this.syncEngine = new Sync4jEngine(Configuration.getConfiguration());
    }

    /**
     * Creates a new instance of SyncSessionHandler with a given session id
     */
    public SyncSessionHandler(String sessionId) {
        this();
        this.sessionId = sessionId;
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Returns true if the sessione has been authenticated.
     *
     * @return true if the sessione has been authenticated, false otherwise
     */
    public boolean isAuthenticated() {
        return (syncState != null) && (syncState.authenticationState == AUTH_AUTHENTICATED);
    }

    /**
     * Returns true if the session has not been authenticated because of an
     * expired account.
     *
     * @return true if the account is expired
     */
    public boolean isAccountExpired() {
        return syncEngine.getOfficer().isAccountExpired();
    }

    /**
     * Processes the given message. See the class description for more
     * information.
     *
     * @return the response message
     * @param context
     * @param message the message to be processed
     * @throws InvalidCredentialsException
     * @throws ProtocolException
     */
    public SyncML processMessage(SyncML message, MessageProcessingContext context)
    throws ProtocolException, InvalidCredentialsException, ServerFailureException {
        SyncML response = null;

        //
        // Reset the cmdIdGenerator has specified in the spec
        //
        resetIdGenerator();

        //
        // Each time a message is received for a particular session adjust the message ID
        //
        msgIdGenerator.next();

        //
        //  We maintain the message Id from client
        //
        lastMsgIdFromClient = message.getSyncHdr().getMsgID();

        //
        // Initialize the device ID from the client request
        //
        clientDeviceId = message.getSyncHdr().getSource().getLocURI();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("current state: " + getStateName(currentState));
        }

        Chal chal = ProtocolUtil.getStatusChal(message);

        try {
            switch (currentState) {
                case STATE_ERROR: // in case of error you can start a new initialization
                case STATE_START:
                    nextTimestamp = new SyncTimestamp();
                    nextTimestamp.start = System.currentTimeMillis();
                    nextTimestamp.tagClient = String.valueOf(nextTimestamp.start);

                    syncState = new SyncState();

                    syncEngine.setSyncTimestamp(new Date(nextTimestamp.start));

                    //
                    // Set maximum message size
                    //
                    Meta meta = message.getSyncHdr().getMeta();
                    if (meta != null && meta.getMaxMsgSize() != null) {
                        syncState.maxMsgSize = meta.getMaxMsgSize().intValue();
                    }

                    //
                    // read device information from the database. Use this
                    // information with care until the client has been
                    // authenticated.
                    //
                    syncState.device = new Sync4jDevice(clientDeviceId);
                    try {
                        syncEngine.readDevice(syncState.device);
                    } catch (NotFoundException e) {
                        if (log.isLoggable(Level.WARNING)) {
                            log.warning( "Client device '"
                                       + clientDeviceId
                                       + "' not found. Authentication may fail."
                                       );
                        }
                    }

                    syncState.syncMLVerProto =
                                message.getSyncHdr().getVerProto().getVersion();
                    syncEngine.setSyncMLVerProto(syncState.syncMLVerProto);

                    //
                    // Check if the client has sent the credential with the same
                    // authentication type imposed by the server
                    //
                    Cred cred = message.getSyncHdr().getCred();

                    if (!checkClientAuth(cred)) {
                        //
                        // the client uses an authentication type different
                        // from authentication server type
                        //
                        syncState.loggedCredential = null;
                        syncState.authenticationState = AUTH_INVALID_CREDENTIALS;

                        if (cred == null) {
                            syncState.authenticationState =
                                             syncState.AUTH_MISSING_CREDENTIALS;
                        }
                    } else {
                        if (login(message.getSyncHdr().getCred(), clientDeviceId)) {
                            if (log.isLoggable(Level.INFO)) {
                                log.info( syncState.loggedPrincipal.getUsername()
                                        + '/'
                                        + syncState.loggedPrincipal.getDeviceId()
                                        + " logged in."
                                        );
                            }
                            if (syncState.loggedPrincipal.getId() == null) {
                                try {
                                    syncEngine.readPrincipal(syncState.loggedPrincipal);
                                } catch (NotFoundException e) {
                                    if (log.isLoggable(Level.INFO)) {
                                        log.info("Authenticated principal not found:" + syncState.loggedPrincipal);
                                    }
                                    syncState.authenticationState = AUTH_INVALID_CREDENTIALS;
                                    syncState.loggedCredential = null;
                                }
                            }

                            //
                            // If the client is authenticated and the authentication
                            // type is MD5, the given client next nonce must be stored
                            // to be used at the next session.
                            //
                            if (chal != null) {
                                if (Cred.AUTH_TYPE_MD5.equals(chal.getType())) {
                                    syncState.device.setServerNonce(Base64.decode(chal.getNextNonce().getValue()));
                                    syncEngine.storeDevice(syncState.device);
                                }
                            }
                        } else {
                            if (log.isLoggable(Level.INFO)) {
                            log.info("Authentication failed for device '"
                                    + clientDeviceId
                                    + "'. Make sure that the client used correct "
                                    + "username and password and that there "
                                    + "is a principal associating the user "
                                    + " to the device. "
                                    );
                            }
                            syncState.authenticationState = AUTH_INVALID_CREDENTIALS;
                            syncState.loggedCredential = null;
                        }
                    }

                    moveTo(STATE_PKG1_RECEIVING);

                case STATE_PKG1_RECEIVING:

                    response = processInitSyncMapMessage(message);

                    if (!isAuthenticated()) {
                        if (message.isLastMessage()) {
                            response.setLastMessage();
                        }
                        moveTo(STATE_START);
                    } else {
                        if (message.isLastMessage()) {
                            moveTo(STATE_PKG3_RECEIVING);
                            if (syncState.cmdCache1.size() == 0) {
                                response.setLastMessage();
                            }
                        }
                    }

                    break;

                case STATE_PKG3_RECEIVING:
                    response = processInitSyncMapMessage(message);

                    if (message.isLastMessage()) {
                        moveTo(STATE_PKG3_RECEIVED);
                        if (syncState.cmdCache3.size() == 0) {
                            response.setLastMessage();
                        }
                    }

                    break;

                case STATE_PKG3_RECEIVED:
                        response = processInitSyncMapMessage(message);

                        if (syncState.cmdCache3.size() == 0) {
                            response.setLastMessage();
                            moveTo(STATE_PKG1_RECEIVING);
                        }

                    break;

                default:
                    logout();
                    throw new ProtocolException("Illegal state: " + currentState);
            }

            if ((currentState != STATE_ERROR)          &&
                (currentState != STATE_START)          &&
                response.isLastMessage()               &&
                (ProtocolUtil.noMoreResponse(response))  ) {

                nextTimestamp.end = System.currentTimeMillis();
                moveTo(STATE_END);
            }

        } catch (ProtocolException e) {
            log.throwing(getClass().getName(), "processMessage", e);
            moveTo(STATE_ERROR);
            throw e;
        } catch (Throwable t) {            
            log.throwing(getClass().getName(), "processMessage", t);
            moveTo(STATE_ERROR);
            throw new ServerFailureException("Server error", t);
        }

        //
        // Reassign ordered cmdId to all commands in the given list.
        //
        List commands = response.getSyncBody().getCommands();
        ProtocolUtil.updateCmdId(commands);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("About returning message: " + Util.toXML(response));
        }

        return response;
    }

    /**
     * Check if the client authentication type is equals to the server
     * authetication type.
     *
     * @param cred the client credential
     *
     * @return true if client and server auth type match, false otherwise
     */
    private boolean checkClientAuth(Cred cred) {

        //
        // Set the client authentication that will be used in the other part
        // of the code too
        //
        Officer officer = syncEngine.getOfficer();
        clientAuth = officer.getClientAuth();

        if (cred == null) {
            syncState.authenticationState = syncState.AUTH_MISSING_CREDENTIALS;
            return false;
        }

        String clientAuthSent = cred.getType();

        if (clientAuth.equalsIgnoreCase(clientAuthSent)) {
            return true;
        }

        syncState.authenticationState = syncState.AUTH_INVALID_CREDENTIALS;
        return false;
    }

    /**
     * Processes an error condition. This method is called when the error is
     * is not fatal and is manageable at a protocol/session level. This results
     * in a well formed SyncML message with an appropriete error code.
     * <p>
     * Note that the offending message <i>msg</i> cannot be null, meaning that
     * at least the incoming message was a SyncML message. In this context,
     * <i>RepresentationException</i>s are excluded.
     *
     * @param the offending message - NOT NULL
     * @param the exception representing the error condition - NOT NULL
     *
     * @throws sync4j.framework.core.Sync4jException in case of unexpected errors
     *
     * @return the response message
     */
    public SyncML processError(SyncML msg, Throwable error)
    throws Sync4jException {
        SyncHdr msgHeader = msg.getSyncHdr();

        Item[] items = new Item[0];
        int status = StatusCode.SERVER_FAILURE;

        items = new Item[1];
        items[0] = new Item(
                       null, // target
                       null, // source
                       null, // meta
                       new ComplexData(error.getMessage()),
                       false //MoreData
                   );

        if (error instanceof ServerException) {
            status = ((ServerException)error).getStatusCode();
        }

        Status statusCommand = new Status(
                cmdIdGenerator.next()               ,
                msgHeader.getMsgID()                ,
                "0" /* command ref */               ,
                "SyncHdr" /* see SyncML specs */    ,
                new TargetRef(msgHeader.getTarget()),
                new SourceRef(msgHeader.getSource()),
                null /* credential */               ,
                null /* challenge */                ,
                new Data(status)                    ,
                new Item[0]
            );

        String serverURI = syncEngine
                         .getConfiguration()
                         .getServerConfig()
                         .getEngineConfiguration()
                         .getServerURI();

        SyncHdr syncHeader = new SyncHdr (
                                  msgHeader.getVerDTD()                        ,
                                  msgHeader.getVerProto()                      ,
                                  msgHeader.getSessionID()                     ,
                                  msgHeader.getMsgID()                         ,
                                  new Target(msgHeader.getSource().getLocURI()),
                                  new Source(serverURI)                        ,
                                  null  /* response URI */                     ,
                                  false                                        ,
                                  null /* credentials */                       ,
                                  null /* metadata */
                               );

        SyncBody syncBody = new SyncBody(
                                new AbstractCommand[] { statusCommand },
                                true /* final */
                            );

        moveTo(STATE_ERROR);

        return new SyncML(syncHeader, syncBody);
    }

    /**
     * Called by the <i>SessionManager</i> when the session is expired.
     * It logs out the credential and release aquired resources.
     */
    public void expire() {
        logout();
    }

    /**
     * Called to interrupt the processing in case of errors depending on
     * extenal causes (i.e. the transport). The current implementation just move
     * the session state to the error state.
     * <p>
     * NOTE that the current implementation simply moves the state of the session
     * to <i>STATE_ERROR</i>.
     *
     * @param statusCode the error code
     *
     * @see sync4j.framework.core.StatusCode for valid status codes
     *
     */
    public void abort(int statusCode) {
        moveTo(STATE_ERROR);
    }

    /**
     * Called to permanently commit the synchronization. It does the following:
     * <ul>
     *  <li>persists the <i>last</i> timestamp to the database for the sources
     *      successfully synchronized
     * </ul>
     */
    public void commit() {
        assert (syncState.loggedPrincipal != null);

        LastTimestamp last = null;

        PersistentStore ps = syncEngine.getStore();

        Database[] dbsGen = syncEngine.getDbs();

        for (int i = 0; (dbsGen != null) && (i < dbsGen.length); ++i) {
            if (dbsGen[i].getStatusCode() != StatusCode.OK) {
                //
                // This database is in error. Do not commit it.
                //
                continue;
            }
            last = new LastTimestamp(
                    syncState.loggedPrincipal.getId(),
                    dbsGen[i].getName(),
                    dbsGen[i].getAnchor().getNext(),
                    dbsGen[i].getServerAnchor().getNext(),
                    nextTimestamp.start,
                    nextTimestamp.end
            );

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Commiting database "
                        + dbsGen[i].getName()
                        + " ( "
                        + last
                        + " )"
                );
            }

            try {
                boolean stored = ps.store(last);
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("LastTimeStamp stored: " + stored);
                }
            } catch (Sync4jException e) {
                log.severe("Error in saving persistent data");
                log.throwing(getClass().getName(), "commit", e);
            }
        }
    }

    // --------------------------------------------------------- Private methods

    /**
     * Processes the given message applying initialization processing first,
     * and then synchronization processing.
     *
     * @param message the message to be processed
     *
     * @return the response message
     *
     * @throws ProtocolException in case of protocol errors
     */
    private SyncML processInitSyncMapMessage(SyncML message)
    throws ProtocolException {
        SyncInitialization  init     = null;
        SyncModifications   sync     = null;
        SyncMapping         map      = null;
        SyncML              response = null;
        SyncHdr             header   = null;

        String msgId = msgIdGenerator.current();

        //
        // First of all check if this message contains initialization elements.
        //
        if (ProtocolUtil.isInitMessage(message)) {
            init = processInitMessage(message);
            header = init.getResponseHeader(msgId);
        }

        //
        // If we are still not authenticated, we do not process sync and map
        // commands and we directly return the init response, which will
        // contains proper status for all commands.
        //
        if (!isAuthenticated()) {
            response = init.getResponseMessage(msgId);
        } else {

            //
            // If the client is authenticated and the authentication
            // type is MD5, the given client next nonce must be stored
            // to be used at the next session.
            //
            Chal chal = ProtocolUtil.getStatusChal(message);
            if (chal != null) {
                if (Cred.AUTH_TYPE_MD5.equals(chal.getType())) {
                    syncState.device.setServerNonce(Base64.decode(chal.getNextNonce().getValue()));
                    syncEngine.storeDevice(syncState.device);
                }
            }

            //
            // Then check if the message contains synchronization elements.
            //
            if (ProtocolUtil.isSyncMessage(message)) {
                sync = processSyncMessage(message);
                if (header == null) {
                    header = sync.getResponseHeader(msgId);
                }
            }

            //
            // Then process any incoming Map command
            //
            if (ProtocolUtil.isMapMessage(message) || ProtocolUtil.noMoreResponse(message)) {
                map = processMapMessage(message);
                if (header == null) {
                    header = map.getResponseHeader(msgId);
                }
            }

            //
            // If we do not have a separate initialization, we jump to the
            // state STATE_PKG3_RECEIVING.
            //
            if ((init != null) && (sync != null)) {
                if (message.isLastMessage()) {
                    moveTo(STATE_PKG3_RECEIVED);
                } else {
                    moveTo(STATE_PKG3_RECEIVING);
                }
            }

            //
            // After processing the commands, if any mapping was modified, the
            // change must be persisted
            //
            syncEngine.storeMappings();

            //
            // If client max msg size is too small, do not even try
            //
            AbstractCommand[] cmdToSend = null;
            if (checkMaxMsgSize()) {
                messageSize = sizeCalculator.getSize(header)
                            + sizeCalculator.getSyncBodyOverhead()
                            + sizeCalculator.getRespURIOverhead()
                            ;
                cmdToSend = commandsToSend(init, sync, map);
            } else {
                cmdToSend = new AbstractCommand[1];

                Status status = (Status)init.getResponseCommands(msgId).get(0);

                status.setData(new Data(StatusCode.SYNCHRONIZATION_FAILED));

                cmdToSend[0] = status;

                if (log.isLoggable(Level.INFO)) {
                    log.info( "The MaxMsgSize ("
                            + syncState.maxMsgSize
                            + ") is smaller than the minimum message size "
                            + "supported by the server. The session will abort."
                            );
                }
            }

            try {
                response = new SyncML(
                    header,
                    new SyncBody(cmdToSend, false)
                );
            } catch (RepresentationException e) {
                throw new ProtocolException(e.getMessage());
            }
        }

        return response;
    }

    /**
     * Processes the initialization commands in the given message like if it
     * contained only initialization information. Init commands are:
     * <ul>
     *   <li>SyncHeader (credentials)</li>
     *   <li>Put (client capabilities)</li>
     *   <li>Get (server capabilities)</li>
     *   <li>Alert (database alerting)</li>
     * </ul>
     *
     * @param message the message to be processed
     *
     * @return the SyncInitialization object representing the initialization
     *         request-response associated to the incoming message
     *
     * @throws ProtocolException in the case of violation of the specification
     */
    private SyncInitialization processInitMessage(SyncML message)
    throws ProtocolException {
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Processing the initialization commands");
        }

        try {
            syncInit = new SyncInitialization(message.getSyncHdr() ,
                                              message.getSyncBody());

            //
            // Store the information that will be used in the Syncronization process
            //
            syncState.addClientAlerts(syncInit.getClientAlerts());

            syncInit.setIdGenerator(cmdIdGenerator);

            //
            // Sets the server authentication type so that in the case the client
            // will be challenged accordingly
            //
            syncInit.setClientAuth(clientAuth);

            //
            // If authentication type is MD5 then generate a new NextNonce
            //
            if (clientAuth.equalsIgnoreCase(Cred.AUTH_TYPE_MD5)) {
                NextNonce nonce = ProtocolUtil.generateNextNonce();
                syncInit.setNextNonce(nonce);
                syncState.device.setClientNonce(nonce.getValue());
                syncEngine.storeDevice(syncState.device);
            }

            if (isAuthenticated()) {
                syncInit.setAuthorizedStatusCode(StatusCode.AUTHENTICATION_ACCEPTED);

                syncInit.setClientCapabilitiesRequired(false);

                //
                // If the client gave its capabilities, process them
                //
                processClientCapabilities(syncInit.getClientDeviceInfo());

                //
                // Gets the databases requested for synchronization and for each of
                // them checks if the database exists on the server and if the
                // credential is allowed to synchronize it
                //
                dbs = syncInit.getDatabasesToBeSynchronized(syncState.loggedPrincipal);

                //
                // This will change the status code of the elements of clientDBs to
                // reflect the availability and accessibility of the given databases
                // on the server
                //
                syncEngine.prepareDatabases(syncState.loggedPrincipal, dbs, nextTimestamp);

                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Requested databases: " + Arrays.asList(dbs));
                }

                boolean noDataSource = true; // there are no datasource allowed to synchronize

                ArrayList dbList = new ArrayList();
                for (int i = 0; ((dbs != null) && (i < dbs.length)); ++i) {
                    syncInit.setStatusCodeForCommand(
                            dbs[i].getAlertCommand(),
                            dbs[i].getStatusCode()
                    );

                    if (dbs[i].isOkStatusCode()) {
                        noDataSource = false;
                        syncEngine.addClientSource(
                                new MemorySyncSource(
                                    dbs[i].getName(),
                                    null,
                                    dbs[i].getSource().getLocURI())
                        );

                        Meta m = ((Item)dbs[i].getAlertCommand().getItems().get(0)).getMeta();
                        if (m != null) {
                            Long l = m.getMaxObjSize();
                            if (l != null) {
                                syncState.maxObjSize = l.longValue();
                            } else {
                                syncState.maxObjSize = 0;
                            }
                        }
                        dbList.add(dbs[i]);
                    }
                }
                dbs = (Database[])dbList.toArray(new Database[dbList.size()]);

                syncEngine.setDbs(dbs);

                //
                // Setting the databases to synchronize. This will force the relative
                // <Alert>s to be inserted in the response.
                //
                syncInit.setDatabases(dbs);

                //
                // Setting server capabilities
                //
                syncInit.setServerCapabilities(
                    syncEngine.getServerCapabilities(syncInit.getDTDVersion())
                );
            } else {
                if (isAccountExpired()) {
                    syncInit.setAuthorizedStatusCode(StatusCode.PAYMENT_REQUIRED);
                } else if (syncState.authenticationState == AUTH_MISSING_CREDENTIALS) {
                    syncInit.setAuthorizedStatusCode(StatusCode.MISSING_CREDENTIALS);
                } else if (syncState.authenticationState == AUTH_INVALID_CREDENTIALS) {
                    syncInit.setAuthorizedStatusCode(StatusCode.INVALID_CREDENTIALS);
                } else {
                    syncInit.setAuthorizedStatusCode(StatusCode.FORBIDDEN);
                }
            }

            //
            // Now we set server credentials. If the client challenged the server,
            // we use the requested auth mechanism, otherwise, we use the default
            // server authentication scheme.
            // Not that if credentials are not set in server configuration,
            // getServerCredential returns null, which means that no credentials
            // will be returned to the client.
            //
            Chal chal = getChal(message);

            String serverAuth = syncEngine.getOfficer().getServerAuth();

            if (chal == null) {
                if (serverAuth.equalsIgnoreCase(Cred.AUTH_NONE)) {
                    syncInit.setServerCredentials(null);
                    syncState.serverAuthenticationState = AUTH_AUTHENTICATED;
                } else {
                    if (syncState.serverAuthenticationState != AUTH_ACCEPTED) {
                        Meta meta = new Meta();
                        meta.setType(serverAuth);
                        meta.setNextNonce(new NextNonce(Base64.encode(
                            syncState.device.getServerNonce()))
                            );

                        chal = new Chal(meta);
                        syncInit.setServerCredentials(
                            syncEngine.getServerCredentials(chal, syncState.device)
                        );
                    }
                }
            } else {
                Cred cred = checkServerAuthentication(message);
                if (cred != null) {
                    syncInit.setServerCredentials(cred);
                }
            }

            return syncInit;

        } catch (Sync4jException e) {
            throw new ProtocolException(e);
        }
    }

    /**
     * Check if the client MaxMsgSize is greater than server minmsgsize.
     * If the client MaxMsgSize is smaller than minmsgsize then change
     * the status code of SyncHdr into StatusCode.SYNCHRONIZATION_FAILED.
     *
     * @return true if the client maxmsgsize is ok, false otherwise
     */
    private boolean checkMaxMsgSize() {
        final long minMsgSize = syncEngine
                                .getConfiguration()
                                .getServerConfig()
                                .getEngineConfiguration()
                                .getMinMaxMsgSize();

        if (log.isLoggable(Level.FINEST)) {
            log.finest( "Checking if MaxMsgSize is larger than the minimum "
                      + "size supported by the server ("
                      + minMsgSize
                      + ")"
                      );
        }

        return ((syncState.maxMsgSize == 0) || (syncState.maxMsgSize >= minMsgSize));
    }

    /**
     * Check if are required the server capabilities and check their size.
     * If that size added at size of SyncHdr and SyncHdr Status and all
     * Alert command is more than MaxMsgSize then set Get Status with
     * StatusCode.REQUESTED_ENTITY_TOO_LARGE
     */
    private void checkSizeCapabilities() {
        //
        // If server capabilities are required then check their size in order
        // to establish if it will be possible to send them
        //
        AbstractCommand[] cmds =
            (AbstractCommand[])syncState.cmdCache1.toArray(new AbstractCommand[0]);
        Status statusGet = (Status)ProtocolUtil.filterCommands(
            cmds,
            Status.class,
            Get.COMMAND_NAME
        );

        if (statusGet != null) {
            ArrayList results = ProtocolUtil.filterCommands(
                cmds,
                Results.class
            );

            if (results.size() > 0) {
                //
                // Calculate size Results
                //
                long sizeCap = sizeCalculator.getSyncMLOverhead()
                             + sizeCalculator.getSyncBodyOverhead()
                             + sizeCalculator.getSize((Results)results.get(0));
                if (sizeCap > syncState.maxMsgSize) {
                    if (log.isLoggable(Level.INFO)) {
                        log.info("Capabilities too big!");
                    }
                    statusGet.setData(new Data(StatusCode.REQUESTED_ENTITY_TOO_LARGE));

                    syncState.cmdCache1.remove(results);
                }
            }
        }
    }

    /**
     * Extracts or create a Chal object that represents how server credentials
     * should be created.<br>
     * If the client challenged the server, we must use the requested auth
     * mechanism, otherwise, we use the default server authentication scheme.
     *
     * @param msg the request message
     *
     * @return a Chal object reqpresenting how server credentials should be
     *         formatted.
     */
    private Chal getChal(final SyncML msg) {
        Chal chal = ProtocolUtil.getStatusChal(msg);
        if (chal != null) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Challenged server authentication with scheme " + chal.getType());
            }
        }

        return chal;
    }

    /**
     * Checks if the server authentication succedeed. It throws a
     * ProtocolException if not.
     *
     * @param msg the SyncML message to check
     *
     * @return true if the server should retry authentication, false otherwise
     *
     * @throw ProcotolException if server authentication did not succeed
     */
    private Cred checkServerAuthentication(SyncML msg)
    throws ProtocolException {
        String serverAuth = syncEngine.getOfficer().getServerAuth();
        int headerStatusCode = ProtocolUtil.getHeaderStatusCode(msg);
        if (headerStatusCode == -1) {
            return null;
        }

        if ( (headerStatusCode == StatusCode.INVALID_CREDENTIALS ||
              headerStatusCode == StatusCode.MISSING_CREDENTIALS)) {
            if (
                (serverAuth.equalsIgnoreCase(Cred.AUTH_TYPE_MD5)   ) &&
                (syncState.serverAuthenticationState == AUTH_UNAUTHENTICATED)
               ) {
                   syncState.serverAuthenticationState = AUTH_RETRY_1;
                   return syncEngine.getServerCredentials(getChal(msg), syncState.device);
            } else {
                throw new ProtocolException("Unable to authenticate the client");
            }
        } else if (headerStatusCode == StatusCode.AUTHENTICATION_ACCEPTED) {
            //
            // Authenticated with code 212
            //
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Server logged (code 212)");
        }
            syncState.serverAuthenticationState = AUTH_ACCEPTED;
        } else {
            //
            // Authenticated with code 200
            //
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Server auhenticated (code 200)");
            }
            syncState.serverAuthenticationState = AUTH_AUTHENTICATED;
            return syncEngine.getServerCredentials(getChal(msg), syncState.device);
        }
        return null;
    }

    /**
     * Processes the given synchronization message.
     *
     * @param syncRequest the message to be processed
     *
     * @return the SyncModifications object representing the request-response
     *         under processing
     *
     * @throws ProtocolException in case of errors
     */
    private SyncModifications processSyncMessage(SyncML syncRequest)
    throws ProtocolException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Processing the given synchronization message");
            log.finest("client sources: " + syncEngine.getClientSources());
        }

        try {
            modifications =
                    new SyncModifications(syncRequest.getSyncHdr() ,
                                          syncRequest.getSyncBody(),
                                          syncEngine.getDbs());

            Sync[] syncCommands = modifications.getClientSyncCommands();

            //
            // Check if into Sync command there is a MaxObjSize for LargeObject
            //
            checkMaxObjSizeIntoSync(syncCommands);

            //
            // TBD: verify this!!!!
            //
            if (syncRequest.isLastMessage()) {
                syncEngine.setLastMessage(true);
            }

            List responseCommands = processModifications(modifications);

            //
            // Add Status for Item with large object
            //
            if (syncState.getListStatusCmdNotProcessed() != null) {
                responseCommands.addAll(syncState.getListStatusCmdNotProcessed());
                syncState.clearListStatusCmdNotProcessed();
            }

            if (log.isLoggable(Level.FINEST)) {
                log.finest("responseCommands: " + responseCommands);
            }

            modifications.setIdGenerator(cmdIdGenerator);
            modifications.setFlag(Flags.FLAG_ALL_RESPONSES_REQUIRED);

            Status[] statusSyncs = (Status[])ProtocolUtil.filterCommands(
                responseCommands,
                new String[]{ Status.COMMAND_NAME }
            ).toArray(new Status[0]);

            modifications.setClientModificationsStatus(statusSyncs);

            // .setServerModifications sets all the modification commands
            // required to be send to client from Server

            // cannot block this whole code,as this code may be used
            // by other DBS within same sync but differnt alert code.
            AbstractCommand[] serverModifications =
                    (AbstractCommand[])ProtocolUtil.filterCommands(
                        responseCommands,
                        new String[]{Sync.COMMAND_NAME}
                    ).toArray(new AbstractCommand[0]);

            //
            // The server must send its modifications only after the client sent
            // the last message
            //
            modifications.setServerModifications(serverModifications);

            return modifications;

        } catch (Sync4jException e) {
            throw new ProtocolException(e);
        }
    }

    /**
     * Checks if the given command can be sent in the current message without
     * exceeding maxMsgSize.
     *
     * @param cmd the command to check
     *
     * @return true if the command can be sent, false otherwise
     */
    private boolean checkSize(AbstractCommand cmd) {
        return (syncState.maxMsgSize != 0) &&
               ( (messageSize + sizeCalculator.getCommandSize(cmd)) < syncState.maxMsgSize);
    }

    /**
     * Checks if the given size is available in the current message without
     * exceeding maxMsgSize.
     *
     * @param size the command to check
     *
     * @return true if the command can be sent, false otherwise
     */
    private boolean checkSize(long size) {
        return (syncState.maxMsgSize != 0) &&
               ( (messageSize + size) < syncState.maxMsgSize);
    }

    /**
     * Checks if the given size is available in the current message without
     * exceeding maxMsgSize starting from the given message size.
     *
     * @param size the command to check
     * @param msgSize the message size
     *
     * @return true if the command can be sent, false otherwise
     */
    private boolean checkSize(long size, long msgSize) {
        return (syncState.maxMsgSize != 0) &&
               ( (msgSize + size) < syncState.maxMsgSize);
    }

    /**
     * Check if into Sync command there is a MaxObjSize for LargeObject.
     * If present cache that value into SyncState.
     *
     * @param Sync[] the array of Sync command client
     */
    private void checkMaxObjSizeIntoSync(Sync[] syncCmds) {
        if (syncCmds != null && syncCmds.length > 0) {
            //
            // The MaxObjSize is equals for every Sync command
            //
            if (syncCmds[0].getMeta() != null &&
                syncCmds[0].getMeta().getMaxObjSize() != null
               ) {
                syncState.maxObjSize = syncCmds[0].getMeta().getMaxObjSize().longValue();
            }
        }
    }

    /**
     * Executes the given modifications and for each of them returns a status
     * command. It forwards the execution to the synchronization engine.
     *
     * @param modifications synchronization commands containing the modifications
     *                 that the client requires
     *
     * @return an array of command objects, each containig the result of
     *         a modification or a modification itself
     * @throws Sync4jException
     */
    private List processModifications(SyncModifications modifications)
    throws Sync4jException {
        Sync[] syncCommands = modifications.getClientSyncCommands();

        ArrayList responseCommands = new ArrayList();

        if ((syncCommands == null) || (syncCommands.length == 0)) {
            //
            // Nothing to process!
            //
            return responseCommands;
        }

        SyncHdr    header           = modifications.getSyncHeader();
        String     msgId            = header.getMsgID()            ;
        boolean    headerNoResponse = header.isNoResp()            ;

        syncEngine.setCommandIdGenerator(cmdIdGenerator);

        //
        // First of all prepare the memory sources with the modification commands
        // receveid by the client. Sent the Status commands to check if there
        // is a "Refresh required": in this case is necessary to do a slow sync.
        //

        Status[] statusCommands =
            (Status[])ProtocolUtil.filterCommands(
                                             modifications.getClientCommands(),
                                             Status.class
            ).toArray(new Status[0]);

        //
        // Check if in the given list of commands there is data to add
        // at the previous data.
        //
        if ((syncCommands != null) && (syncCommands.length>0)) {
            checkForReceivedLargeObject(syncCommands);
        }

        //
        // First of all prepare the memory sources with the modification commands
        // receveid by the client
        //
        prepareMemorySources(syncCommands, statusCommands);

        try {
            syncEngine.sync(syncState.loggedPrincipal);
        } catch (Sync4jException e) {
            log.throwing(getClass().getName(), "processModifications", e);

            if (log.isLoggable(Level.SEVERE)) {
                log.severe(e.getMessage());
            }
        }

        //
        // Status code for commands
        //
        if (headerNoResponse == false) {
            //
            // Sync commands
            //
            responseCommands.addAll(statusForSyncs(syncCommands));

            //
            // Status for server-side executed modification commands
            //
            Status[] operationStatus =
                syncEngine.getModificationsStatusCommands(msgId);

            for (int i=0; i<operationStatus.length; ++i) {
                responseCommands.add(operationStatus[i]);
            }
        }

        //
        // SyncCommands sent back to the client
        // No sync sent to client if ONE_WAY_FROM_CLIENT
        //

        Database[] dbs = syncEngine.getDbs();
        ItemizedCommand[] commands = null;
        String uri = null;
        for (int i = 0; (i < dbs.length); ++i) {
            //
            // If the alert represents a client only action, just ignore it
            //
            if (AlertCode.isClientOnlyCode(dbs[i].getMethod())) {
                continue;
            }

            uri = dbs[i].getName();

            SyncOperation[] operations = syncEngine.getSyncOperations(uri);
            commands = syncEngine.operationsToCommands(operations, uri, slow);
            Long noc = isNumberOfChangesSupported()
                     ? new Long(commands.length)
                     : null
                     ;
            responseCommands.add(
                    new Sync(
                            cmdIdGenerator.next(),
                            false,
                            null,
                            dbs[i].getTarget(),
                            dbs[i].getSource(),
                            null,
                            noc,
                            commands
                    )
            );
            syncEngine.resetSyncOperations(uri);
        } // next i

        //
        // And here we can update the server-side LUID-GUID mappings
        //
        syncEngine.updateServerMappings(slow);

        return responseCommands;
    }

    /**
     * Process the completion message from the client check status for sync
     * command and update client mapping
     * @param message
     * @return the response message
     * @throws ProtocolException
     */
    private SyncMapping processMapMessage(SyncML message)
    throws ProtocolException {
        if (log.isLoggable(Level.INFO)) {
            log.info("Handling mapping ...");
        }

        try {
            SyncMapping mapping = new SyncMapping(
                                  message.getSyncHdr(),
                                  message.getSyncBody()
                              );

            mapping.setIdGenerator(cmdIdGenerator);

            if (log.isLoggable(Level.INFO)) {
               log.info("Performing mapping ...");
            }

            String uri = null;

            sync4j.framework.core.Map[] mapCommands = mapping.getMapCommands();
            MapItem[]    mapItems    = null                             ;
            for (int j=0; ((mapCommands != null) && (j<mapCommands.length)); ++j) {
                uri = mapCommands[j].getTarget().getLocURI();
                mapItems =
                    (MapItem[])mapCommands[j].getMapItems().toArray(
                                                            new MapItem[0]);

                for (int i = 0; i < mapItems.length; i++) {
                    MapItem mapItem = mapItems[i];

                    //Adding item properties to the persistent mapping
                    String guid = mapItem.getTarget().getLocURI();
                    String luid = mapItem.getSource().getLocURI();
                    syncEngine.updateMapping(uri, guid, luid);
                }
            }

            return mapping;

        } catch (Sync4jException e) {
            throw new ProtocolException(e);
        }
    }

    /**
     * Makes a state transition. Very simple implementation at the moment: it
     * changes the value of <i>currentState</i> to the given value.
     *
     * @param state the new state
     */
    private void moveTo(int state) {
        if (log.isLoggable(Level.FINEST)) {
            log.finest("moving to state " + getStateName(state));
        }
        currentState = state;
    }

    /**
     * Prepares the memory sources with the modification commands receveid
     * by the client.
     * <p>
     * Note that if the requested synchronization is a slow sync, the items are
     * inserted as "existing" items, regardless the command they belong to.
     * (maybe this will change as soon as the specification becomes clearer)
     * If there is no alert for slow sync, is need to check if exist a status
     * with "Refresh required" (code 508)
     *
     * @param syncCommands the commands used to prepare the source
     * @param statusCommands the status commands sent by the client
     *
     */
    private void prepareMemorySources(Sync[] syncCommands, Status[] statusCommands) {

        List sources = syncEngine.getClientSources();

        //
        // For efficiency: put the databases in a HashMap
        //
        HashMap dbMap = new HashMap();
        for (int i=0; ((syncEngine.getDbs() != null) && (i<syncEngine.getDbs().length)); ++i) {
            dbMap.put((syncEngine.getDbs())[i].getName(), (syncEngine.getDbs())[i]);
        }

        //
        // First of all prepare the memory sources with the modification commands
        // receveid by the client
        //
        int method;
        slow = false;
        Target target = null;
        MemorySyncSource mss = null;
        AbstractCommand[] modifications = null;
        for (int i = sources.size(); i > 0; --i) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Preparing "
                          + syncEngine.getClientSources().get(i - 1)
                          + " with "
                          + Arrays.asList(syncCommands)
                );
            }

            mss = (MemorySyncSource) sources.get(i - 1);

            String  uri = mss.getSourceURI();

            modifications = new AbstractCommand[0];
            ArrayList alModifications = new ArrayList();

            //
            // The client device could send more Sync commands for the same source
            //
            for (int j = 0; ((syncCommands != null) && (j < syncCommands.length)); ++j) {
                target = syncCommands[j].getTarget();
                if ((target != null) && (uri.equals(target.getLocURI()))) {
                    if (syncCommands[j].getCommands() != null) {
                        alModifications.addAll(syncCommands[j].getCommands());
                    }
                }
            }

            modifications = (AbstractCommand[])alModifications.toArray(new AbstractCommand[0]);

            method = ((Database)dbMap.get(uri)).getMethod();
            slow = ((method == AlertCode.SLOW) || (method == AlertCode.REFRESH_FROM_SERVER));

            if (!slow) {
                String targetRef = null;
                String statusCode = null;
                ArrayList targetRefs = null;
                for (int k=0; statusCommands!=null && k<statusCommands.length; k++) {
                    targetRefs = statusCommands[k].getTargetRef();
                    if (targetRefs == null || targetRefs.size() == 0) {
                        continue;
                    }
                    targetRef = ((TargetRef)targetRefs.get(0)).getValue();
                    if ((targetRef != null) && (uri.equals(targetRef))) {
                        statusCode = statusCommands[k].getData().getData();
                        if (statusCode.equals("" + StatusCode.REFRESH_REQUIRED)) {
                            slow = true;
                            ((Database)dbMap.get(uri)).setStatusCode(StatusCode.REFRESH_REQUIRED);
                        }
                        break;
                    }
                }
            }

            prepareMemorySource(mss, modifications, slow);
        }
    }

    /**
     * Prepares a source that represents the image of the client database. This
     * is done combining the existing client mapping with the modifications sent
     * by the client. Than this source can be compared with the server view of
     * the database.
     *
     * @param source the e source to prepare
     * @param commands the client modifications
     * @param slowSync true if the preparation is for a slow sync, false
     *                    otherwise
     *
     */
    private void prepareMemorySource(MemorySyncSource  source  ,
                                     AbstractCommand[] commands,
                                     boolean           slowSync) {
        ArrayList deleted  = new ArrayList();
        ArrayList created  = new ArrayList();
        ArrayList updated  = new ArrayList();

        ClientMapping guidluid = null;

        //
        // First of all, in the case of fast sync, the items already mapped are
        // added as existing items (with GUID as key).
        // In the case of slow sync, the mappings are cleared so that they
        // won't generate conflicts.
        //
        guidluid = (ClientMapping)syncEngine.getMapping(
                       syncState.loggedPrincipal,
                       source.getSourceURI(),
                       slowSync
                   );

        String name = null;
        for (int i = 0; ((commands != null) && (i < commands.length)); ++i) {
            if (commands[i] == null) {
                continue;
            }
            name = commands[i].getName();
            if (slowSync && !Delete.COMMAND_NAME.equals(name)) {
                List items = Arrays.asList(
                                 syncEngine.itemsToSyncItems(
                                     source,
                                     (ModificationCommand)commands[i],
                                     SyncItemState.SYNCHRONIZED,
                                     nextTimestamp.start
                                 )
                             );
                updated.addAll(items);

                continue;
            }
            if (Add.COMMAND_NAME.equals(commands[i].getName())) {
                created.addAll(
                        Arrays.asList(
                                syncEngine.itemsToSyncItems(
                                    source,
                                    (ModificationCommand)commands[i],
                                    SyncItemState.NEW,
                                    nextTimestamp.start
                                )
                        )
                );
                continue;
            }

            if (Delete.COMMAND_NAME.equals(commands[i].getName())) {
                deleted.addAll(
                        Arrays.asList(
                                syncEngine.itemsToSyncItems(
                                    source,
                                    (ModificationCommand)commands[i],
                                    SyncItemState.DELETED,
                                    nextTimestamp.start
                                )
                        )
                );
                continue;
            }

            if (Replace.COMMAND_NAME.equals(commands[i].getName())) {
                updated.addAll(
                        Arrays.asList(
                                syncEngine.itemsToSyncItems(
                                    source,
                                    (ModificationCommand)commands[i],
                                    SyncItemState.UPDATED,
                                    nextTimestamp.start
                                )
                        )
                );
                continue;
            }
        }

        source.initialize(deleted, created, updated);

        if (syncEngine.isLastMessage()) {
            source.setExistingItems(createItemsFromMapping(source));
        }
    }

    /**
     * Create and return the status commands for the executed &lt;Sync&gt;s.
     *
     * @param syncCommands the Sync commands
     *
     * @return the status commands in a List collection
     */
    private List statusForSyncs(Sync[] syncCommands) {
        ArrayList ret = new ArrayList();

        String uri = null;
        Target target = null;
        Source source = null;
        TargetRef targetRef = null;
        SourceRef sourceRef = null;
        int statusCode = StatusCode.OK;
        String statusMessage = "";
        Item[] items = null;

        for (int i = 0; (  (syncCommands     != null )
                        && (i < syncCommands.length)); ++i) {

            target = syncCommands[i].getTarget();
            source = syncCommands[i].getSource();

            //
            // A Sync command can be empty....
            //
            uri = (target==null) ? null : target.getLocURI();
            if ((uri == null) || (syncEngine.getClientSource(uri) != null)) {
                statusCode    = syncEngine.getClientSourceStatus(uri);
                statusMessage = syncEngine.getClientStatusMessage(uri);
            } else {
                statusCode = StatusCode.NOT_FOUND;
                statusMessage = null;
            }

            targetRef = (target == null) ? null : new TargetRef(uri);

            sourceRef = (source == null) ? null : new SourceRef(syncCommands[i].getSource());

            //
            // If there is a status message to report, add a new Item to the
            // Status command. Otherwise, no item is appended
            //
            if (statusMessage != null) {
                items = new Item[] {
                            new Item(null, null, null, new ComplexData(statusMessage), false)
                        };
            } else {
                items = new Item[0];
            }

            ret.add(
                new Status(
                        cmdIdGenerator.next(),
                        lastMsgIdFromClient,
                        syncCommands[i].getCmdID().getCmdID(),
                        syncCommands[i].COMMAND_NAME,
                        targetRef,
                        sourceRef,
                        null,
                        null,
                        new Data(statusCode),
                        items
                )
            );
        } // next i

        return ret;
    }

    /**
     * Checks that the credentials of the given message are allowed to start a
     * session.
     *
     * @param credential the message
     * @param deviceId the deviceId
     */
    private boolean login(Cred credential, String deviceId) {
        //
        // May be the credential is already logged in...
        //
        logout();

        if (credential == null) {
            syncState.authenticationState = AUTH_MISSING_CREDENTIALS;
            return false;
        }

        Sync4jPrincipal p = Sync4jPrincipal.fromCredential(
            credential.getData(),
            credential.getType(),
            deviceId
        );
        Authentication auth = credential.getAuthentication();
        auth.setDeviceId(deviceId);
        auth.setSyncMLVerProto(syncState.syncMLVerProto);
        NextNonce nn = new NextNonce(syncState.device.getClientNonce());
        auth.setNextNonce(nn);

        if (syncEngine.login(credential)) {
            if (clientAuth.equalsIgnoreCase(Cred.AUTH_TYPE_MD5)) {
                p.setUsername(credential.getAuthentication().getUsername());
            }

            //
            // The logged principal is find only after the authentication process
            //
            p.setId(credential.getAuthentication().getPrincipalId());

            if (syncEngine.authorize(p, SecurityConstants.RESOURCE_SESSION)) {

                syncState.loggedCredential = credential;
                syncState.loggedPrincipal  = p         ;

                syncState.authenticationState = AUTH_AUTHENTICATED;

                return true;
            }
        }
        return false;
    }

    /**
     * Logs out the logged in credential
     */
    private void logout() {
        if (isAuthenticated()) {
            syncEngine.logout(syncState.loggedCredential);
        }
        syncState.authenticationState = AUTH_INVALID_CREDENTIALS;
        syncState.loggedCredential = null;
        syncState.loggedPrincipal  = null;
    }

    /**
     * Called by the <i>SyncBean</i> when the container release the session.
     * It commit the change to the DB, logs out the credential and
     * release aquired resources.
     */
    public void endSession() {
        commit();
        logout();
    }

    private String getStateName(int state) {
        String stateName = "STATE_UNKNOWN";

        switch (state) {
            case STATE_START                      : stateName = "STATE_START"                     ; break;
            case STATE_END                        : stateName = "STATE_END"                       ; break;
            case STATE_PKG1_RECEIVING             : stateName = "STATE_PKG1_RECEIVING"            ; break;
            case STATE_PKG1_RECEIVED              : stateName = "STATE_PKG1_RECEIVED"             ; break;
            case STATE_PKG3_RECEIVING             : stateName = "STATE_PKG3_RECEIVING"            ; break;
            case STATE_PKG3_RECEIVED              : stateName = "STATE_PKG3_RECEIVED"             ; break;
            case STATE_ERROR                      : stateName = "STATE_ERROR"                     ; break;
            default                               : stateName = "STATE_UNKNOWN"                   ; break;
        }

        return stateName;
    }

    public SyncState getSyncState() {
        return this.syncState;
    }

    /**
     * Checks if in the given list of commands there is data to add to
     * previous data. The command with the data to add is the first in the list.
     * If there is a command, merges the previous data with the data contained in
     * the command.
     *
     * @param syncs the command list to check
     *
     * @throws ProtocolException if the list not contains a valid Results
     */
    private void checkForReceivedLargeObject(Sync[] syncs) {
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Checking if there are data to add to previous data");
        }

        Item lo = null;

        if (syncState.receivedLargeObject != null) {
            Item item =
                ProtocolUtil.getSyncItem(syncs, syncState.receivedLargeObject);

            if (item == null) {
                if (log.isLoggable(Level.WARNING)) {
                    log.warning("The end of data for a chuncked objects has NOT been received!");
                }

                createAlert(AlertCode.NO_END_OF_DATA);

                syncState.receivedLargeObject       = null;
                syncState.sizeOfReceivedLargeObject = null;
                syncState.syncLocURI                = null;
                syncState.itemLocURI                = null;

                return;
            } else {
                if (!item.isMoreData()) {
                    //
                    // This is the last chunk
                    //
                    syncState.receivedLargeObject = null;
                }
            }
        }

        if (syncState.receivedLargeObject == null) {
            int i = 0;
            for (i=0; (lo == null) && (i<syncs.length); ++i) {
                lo = ProtocolUtil.getLargeObject(syncs[i].getCommands());
            }

            if (lo == null) {
                //
                // no LO found
                //
                return;
            }

            String loURI = (lo.getSource() != null)
                         ? lo.getSource().getLocURI()
                         : lo.getTarget().getLocURI()
                         ;

            syncState.receivedLargeObject = syncs[i-1].getTarget().getLocURI()
                                          + '/'
                                          + loURI
                                          ;
            return;
        }
    }

    private void createAlert(int alertCode) {
        if (syncState.receivedLargeObject == null) {
            return;
        }

        int p = syncState.receivedLargeObject.lastIndexOf('/');

        if (p<0) {
            return;
        }

        String uri = syncState.receivedLargeObject.substring(p+1);

        Item item = new Item(new Target(uri), null, null, null, false);
        ArrayList items = new ArrayList(1);
        items.add(item);

        syncState.cmdCache3.add(
            new Alert(
                cmdIdGenerator.next(),
                false,
                null,
                alertCode,
                (Item[])(items.toArray(new Item[0]))
            )
        );
    }

    /**
     * Returns an array of AbstractCommand containing the commands to send in
     * the response message taking into account engine state, sent commands and
     * max message size.
     *
     * @param init a SyncInitialization instance
     * @param sync a SyncModifications instance
     * @param map  a ClientMapping instance
     *
     * @return an array of AbstractCommand containing the commands to send in
     *         the response message
     *
     * @throws ProtocolException in case of protocol errors
     */
    private AbstractCommand[] commandsToSend(SyncInitialization  init,
                                             SyncModifications   sync,
                                             SyncMapping         map )
    throws ProtocolException {

        //
        // Step 1: cache all commands to be sent on the proper array list
        //
        if (init != null) {
            syncState.cmdCache1.addAll(init.getResponseCommands(msgIdGenerator.current()));
        }

        if (sync != null) {
            List commands = sync.getResponseCommands(msgIdGenerator.current());
            List status = ProtocolUtil.filterCommands(commands, new String[] { Status.COMMAND_NAME });
            syncState.cmdCache1.addAll(status);
            commands.removeAll(status);

            syncState.cmdCache3.addAll(commands);
            //
            // Note: this could be done more efficiently when inserting the
            // commands (instead of using addAll...
            //
            ProtocolUtil.mergeSyncCommands(syncState.cmdCache3);
        }

        if (map != null) {
            List commands = map.getResponseCommands(msgIdGenerator.current());
            List status = ProtocolUtil.filterCommands(commands, new String[] { Status.COMMAND_NAME });
            syncState.cmdCache1.addAll(status);
            commands.removeAll(status);
            syncState.cmdCache3.addAll(commands);
        }

        ProtocolUtil.removeHeaderStatus(syncState.cmdCache1);

        //
        // Step 2: first of all we have to send all PKG#1 responses, status
        //         first and then Alerts and Results
        //
        List              cache    = null; // which cache are we using?
        AbstractCommand[] allCmds  = null;
        AbstractCommand[] statuses = null;
        AbstractCommand[] commands = null;

        cache = new ArrayList();

        //
        // Modification commands can be sent only after PKG1 has been completely
        // received
        //
        cache.addAll(syncState.cmdCache1);
        if (currentState == STATE_PKG3_RECEIVED) {
            cache.addAll(syncState.cmdCache3);
        }

        allCmds = (AbstractCommand[])
                    cache.toArray(new AbstractCommand[cache.size()]);

        ArrayList statusList  = ProtocolUtil.filterCommands(allCmds, Status.class);
        ArrayList commandList = ProtocolUtil.inverseFilterCommands(allCmds, Status.class);

        statuses = ProtocolUtil.sortStatusCommand(statusList);
        commands =
                (AbstractCommand[])commandList.toArray(new AbstractCommand[commandList.size()]);

        //
        // If we would have to send server capabilities, but they are too big,
        // we remove them from the commands to send and we set a proper state.
        //
        checkSizeCapabilities();

        //
        // Step 3: select the commands we can send without exceding maxMsgSize
        //
        commandList = new ArrayList();
        for (int i=0; i<statuses.length; ++i) {
            if (!checkSize(statuses[i])) {
                return (AbstractCommand[])commandList.toArray(new AbstractCommand[0]);
            }

            messageSize += sizeCalculator.getCommandSize(statuses[i]);
            commandList.add(statuses[i]);
            syncState.cmdCache1.remove(statuses[i]);
        }

        for (int i=0; i<commands.length; ++i) {
            if (!checkSize(commands[i])) {
                if (commands[i] instanceof Sync) {
                    Sync s = splitSync((Sync)commands[i]);
                    commandList.add(s);
                    messageSize += sizeCalculator.getCommandSize(s);
                }

                return (AbstractCommand[])commandList.toArray(new AbstractCommand[0]);
            }

            messageSize += sizeCalculator.getCommandSize(commands[i]);
            commandList.add(commands[i]);
            syncState.cmdCache1.remove(commands[i]);
            syncState.cmdCache3.remove(commands[i]);
        }

        return (AbstractCommand[])commandList.toArray(new AbstractCommand[0]);
    }

    /**
     * Given a Sync command, a new Sync command is created with no more items
     * then the ones that fit in the message. The item inserted into the new
     * command are removed from the input command.
     *
     * @param cmd the Sync commmand
     *
     * @return a new Sync command with only the items that fit into the message size
     *
     */
    private Sync splitSync(Sync cmd) {
        ArrayList commands = cmd.getCommands();

        //
        // First of all remove all commands that will never be sent; these are:
        //  - Items bigger then MaxObjSize (if specfied)
        //  - Items bigger then MaxMsgSize-MSG_OVERHEAD if the client does not
        //    support large object
        //
        boolean lo = isLargeObjectSupported();
        if (lo && (syncState.maxObjSize > 0)) {
            removeTooBigCommands(commands, syncState.maxObjSize);
        } else if (!lo) {
            removeTooBigCommands(commands, syncState.maxMsgSize - sizeCalculator.getMsgSizeOverhead());
        }

        Sync newSync = new Sync( cmdIdGenerator.next(),
                                 cmd.isNoResp(),
                                 cmd.getCred(),
                                 cmd.getTarget(),
                                 cmd.getSource(),
                                 cmd.getMeta(),
                                 null,
                                 new AbstractCommand[0]);

        ArrayList newCommands = new ArrayList();

        long syncSize = sizeCalculator.getCommandSize(newSync);
        long tmpMessageSize = messageSize;

        int howManyCommands = commands.size();

        Iterator i = commands.iterator();
        while (i.hasNext()) {
            AbstractCommand c = (AbstractCommand)i.next();

            if (!checkSize(syncSize + sizeCalculator.getCommandSize(c), tmpMessageSize)) {
                break;
            }

            newCommands.add(c);
            tmpMessageSize += sizeCalculator.getCommandSize(c);
        }

        commands.removeAll(newCommands);

        //
        // If no commands were sent, let's send them as large objects (if
        // supported by the client).
        //
        if (lo && (howManyCommands > 0) && (howManyCommands == commands.size())) {
            ItemizedCommand c = (ItemizedCommand)commands.get(0);
            newCommands.add(nextLOChunk(cmd, c, tmpMessageSize));
            if (((Item)c.getItems().get(0)).getData().getData().length() == 0) {
                commands.remove(c);
            }
        }

        //
        // If number of changes was specified, adjust it to the new values
        //
        if (cmd.getNumberOfChanges() != null) {
            cmd.setNumberOfChanges(new Long(commands.size()));
            newSync.setNumberOfChanges(new Long(newCommands.size()));
        }

        newSync.setCommands((AbstractCommand[])newCommands.toArray(new AbstractCommand[0]));

        return newSync;
    }

    /**
     * Checks if the current mesasge is the last of the package. A message is
     * the last of the package if its corresponding command cache is empty and
     * the state is a "RECEIVED" state.
     *
     * @return true if the message is the last of the package, false otherwise
     */
    private boolean isLastMessage() {
        return ((currentState == STATE_PKG1_RECEIVED) && (syncState.cmdCache1.size() == 0))
            || ((currentState == STATE_PKG3_RECEIVED) && (syncState.cmdCache3.size() == 0))
            ;
    }

    /**
     * Processes the client capabilities setting the device object accordingly.
     *
     * @param devInfo NULL - if null, no client caps were given
     */
    private void processClientCapabilities(DevInf devInfo) {
        if (devInfo != null) {
            syncState.devInf = devInfo;
        }
    }

    /**
     * Removes the commands bigger then the given threshold from the given list.
     *
     * @param commands the command list
     * @param size the threshold size
     */
    private void removeTooBigCommands(List commands, long size) {
        ArrayList remove = new ArrayList();

        Iterator i = commands.iterator();
        while(i.hasNext()) {
            ItemizedCommand c = (ItemizedCommand)i.next();
            Item item = (Item)c.getItems().get(0);
            Data d = item.getData();
            if ((d != null) && (d.getData().length() > size)) {
                if (log.isLoggable(Level.WARNING)) {
                    log.warning( "Item "
                               + item.getSource().getLocURI()
                               + " is bigger than the maximum allowed size ("
                               + size
                               + "). It is removed from the list of commands to send."
                               );
                }
                remove.add(c);
            }
        }
        commands.removeAll(remove);
    }

    /**
     * Splits the data of the given command. The new chunk of data will be
     * returned in newly created command; data sent are removed from cmd. The
     * first time, a Meta with the object size is stored in cmd so that we can
     * detect when this is the first chunk or not.
     *
     * @param sync the Sync command containing cmd
     * @param cmd the command containing the object to split
     * @param msgSize the size of the message that will include the chunk
     *
     * @return boolean true if the data is splitted, false otherwise
     */
    private ItemizedCommand nextLOChunk(Sync sync, ItemizedCommand cmd, long msgSize) {
        if (!(cmd instanceof Add) && !(cmd instanceof Replace)) {
            //
            // This should never happen!!!
            //
            return null;
        }

        Item item = (Item)cmd.getItems().get(0);

        Data data = item.getData();
        String dataValue = data.getData();
        int dataLength = dataValue.length();

        ItemizedCommand chunk = null;

        Item chunkItem = new Item(item.getTarget()   ,
                                  item.getSource()   ,
                                  null               ,
                                  new ComplexData(""),
                                  false              );

        String chunkURI = null;
        if (cmd instanceof Add) {
            chunk = new Add(cmd.getCmdID()          ,
                            cmd.isNoResp()          ,
                            cmd.getCred()           ,
                            cmd.getMeta()           ,
                            new Item[] { chunkItem });
            chunkURI = chunkItem.getSource().getLocURI();
        } else if (cmd instanceof Replace) {
            chunk = new Replace(cmd.getCmdID()          ,
                                cmd.isNoResp()          ,
                                cmd.getCred()           ,
                                cmd.getMeta()           ,
                                new Item[] { chunkItem });
            chunkURI = chunkItem.getTarget().getLocURI();
        }

        String loURI = sync.getSource().getLocURI()
                     + '/'
                     + chunkURI
                     ;
        if (!loURI.equals(syncState.sendingLOURI)) {
            Meta m = cmd.getMeta();
            if (m == null) {
                m =  new Meta();
            }

            Meta chunkMeta = new Meta();
            chunkMeta.setType(m.getType());
            chunkMeta.setSize(new Long(dataLength));

            chunk.setMeta(chunkMeta);

            syncState.sendingLOURI = loURI;
        } else {
            chunk.setMeta(null);
        }

        //
        // calculate the chunk length
        //
        int chunkLength =
            (int)(syncState.maxMsgSize - (msgSize + sizeCalculator.getMsgSizeOverhead()));
        if (dataLength < chunkLength) {
            chunkItem.getData().setData(dataValue);
            chunkItem.setMoreData(Boolean.FALSE);
            item.getData().setData("");
        } else {
            chunkItem.getData().setData(dataValue.substring(0, chunkLength));
            chunkItem.setMoreData(Boolean.TRUE);
            item.getData().setData(dataValue.substring(chunkLength));
        }

        return chunk;
    }

    /**
     * Does the client support large object? The returned value is the one
     * in the device object if not differently specified by the client in the
     * device capabilities. If so, the information in the DevInf object
     * overwrites the one stored into the database.
     */
    private boolean isLargeObjectSupported() {
        return
            (syncState.maxObjSize > 0)
            ||
            ((syncState.devInf != null) && syncState.devInf.isSupportLargeObjs())
            ||
            syncState.device.getCapabilities().isSupportLargeObject()
            ;
    }

    /**
     * Does the client support large object? The returned value is the one
     * in the device object if not differently specified by the client in the
     * device capabilities. If so, the information in the DevInf object
     * overwrites the one stored into the database.
     */
    private boolean isNumberOfChangesSupported() {
        if (syncState.devInf != null) {
            return syncState.devInf.isSupportNumberOfChanges();
        }

        return syncState.device.getCapabilities().isSupportNumberOfChanges();
    }

    /**
     * Recreates the device sync sources from the LUID-GUID mapping.
     *
     * @param source the source the items belong to
     *
     * @return a List representing the items on the device database
     */
    private List createItemsFromMapping(SyncSource source) {
        ClientMapping m = syncEngine.getMapping(source.getSourceURI());

        ArrayList items = new ArrayList();

        if (m == null) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe( "Client mapping not found for "
                          + source.getSourceURI()
                          + '!'
                          );
                return items;
            }
        }

        java.util.Map map = m.getMapping();

        Iterator i = map.keySet().iterator();
        Object key = null;
        while (i.hasNext()) {
            key = i.next();
            items.add(
                new SyncItemImpl(source, key, map.get(key), SyncItemState.SYNCHRONIZED)
            );
        }

        return items;
    }
}
