/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.server.store;

import java.util.ArrayList;
import java.sql.*;

import sync4j.framework.tools.Base64;
import sync4j.framework.tools.DBTools;
import sync4j.framework.server.store.*;
import sync4j.framework.server.Sync4jDevice;
import sync4j.framework.server.Capabilities;

import sync4j.server.tools.IdSpaceGenerator;

/**
 * This is the store for device information in the Sync4j engine.
 * Currently it persistes the following classes:
 * <ul>
 * <li>sync4j.server.engine.Sync4jDevice</li>
 * </ul>
 *
 * @author  Stefano Fornari @ Funambol
 *
 * @version $Id: DevicePersistentStore.java,v 1.15 2005/07/25 15:26:56 nichele Exp $
 *
 */
public class DevicePersistentStore
extends BasePersistentStore
implements PersistentStore, java.io.Serializable {

    // --------------------------------------------------------------- Constants

    private static final String DEFAULT_SERVER_PASSWORD = "sync4j";

    private static final String OPT_INSERT              = "INSERT";

    public static final int SQL_SELECT_ALL_DEVICES           = 0;
    public static final int SQL_GET_DEVICE                   = 1;
    public static final int SQL_INSERT_DEVICE                = 2;
    public static final int SQL_UPDATE_DEVICE                = 3;
    public static final int SQL_DELETE_DEVICE                = 4;
    public static final int SQL_COUNT_DEVICES                = 5;

    public static final String NS_DEVICE    = "device"   ;

    // -------------------------------------------------------------- Properties

    protected String[] sql = null;

    public void setSql(String[] sql) {
        this.sql = sql;
    }

    public String[] getSql() {
        return this.sql;
    }

    // ------------------------------------------------------------ Private data
    private static final String SEARCH_COUNT_DEVICES    = "SCD";

    // ------------------------------------------------------------ Constructors
    // ---------------------------------------------------------- Public methods

    /**
     * Delete the device from the data store.
     *
     * @param o The object that must be a device for this method
     *
     * @throws PersistentException in case of error reading the data store
     */
    public boolean delete(Object o) throws PersistentStoreException {
        if (o instanceof Sync4jDevice) {
            Sync4jDevice d = (Sync4jDevice) o;

            Connection conn = null;
            PreparedStatement stmt = null;

            try {
                conn = dataSource.getConnection();

                stmt = conn.prepareStatement(sql[SQL_DELETE_DEVICE]);
                stmt.setString(1, d.getDeviceId());
                stmt.executeUpdate();

                return true;
            } catch (SQLException e) {
                throw new PersistentStoreException("Error deleting the device " + d, e);
            } finally {
                DBTools.close(conn, stmt, null);
            }
        }
        return false;
    }

    public boolean store(Object o, String operation) throws PersistentStoreException {
        return false;
    }

    /**
     * Insert or Update the device from the data store.
     *
     * @param o The object that must be a device for this method
     *
     * @throws PersistentException in case of error reading the data store
     */
    public boolean store(Object o) throws PersistentStoreException {
        if (o instanceof Sync4jDevice) {
            Sync4jDevice d = (Sync4jDevice) o;

            Connection conn = null;
            PreparedStatement stmt = null;
            int n = 0;

            try {
                conn = dataSource.getConnection();

                String serverPwd = DEFAULT_SERVER_PASSWORD;

                if (d.getServerPassword() != null) {
                    serverPwd = d.getServerPassword();
                }

                if (d.getDeviceId() != null && !d.getDeviceId().equals("")) {
                    stmt = conn.prepareStatement(sql[SQL_UPDATE_DEVICE]);

                    stmt.setString(1, d.getDescription());
                    stmt.setString(2, d.getType());
                    byte[] nonce = d.getClientNonce();
                    stmt.setString(3, ((nonce != null) ? new String(Base64.encode(nonce)) : null));
                    nonce = d.getServerNonce();
                    stmt.setString(4, ((nonce != null) ? new String(Base64.encode(nonce)) : null));
                    stmt.setString(5, serverPwd);
                    stmt.setString(6, d.getTimeZone());
                    stmt.setString(7, (d.getConvertDate() ? "Y" : "N" ));
                    stmt.setString(8, d.getCharset());
                    stmt.setString(9, d.getDeviceId());

                    n = stmt.executeUpdate();
                    stmt.close(); stmt = null;
                }

                if (n == 0) {
                    // The first time!!!
                    if (d.getDeviceId() == null || d.getDeviceId().equals("")) {
                        int deviceId = getDeviceId();
                        d.setDeviceId("" + deviceId);
                    }

                    stmt = conn.prepareStatement(sql[SQL_INSERT_DEVICE]);
                    stmt.setString(1, d.getDeviceId());
                    stmt.setString(2, d.getDescription());
                    stmt.setString(3, d.getType());
                    byte[] nonce = d.getClientNonce();
                    stmt.setString(4, ((nonce != null) ? new String(Base64.encode(nonce)) : null));
                    nonce = d.getServerNonce();
                    stmt.setString(5, ((nonce != null) ? new String(Base64.encode(nonce)) : null));

                    stmt.setString(6, serverPwd);

                    //
                    // Set id_caps = 2 like default value
                    //
                    stmt.setInt(7, 2);
                    stmt.setString(8, d.getTimeZone());
                    stmt.setString(9, (d.getConvertDate() ? "Y" : "N" ));
                    stmt.setString(10, d.getCharset());

                    stmt.executeUpdate();
                }
            } catch (SQLException e) {
                throw new PersistentStoreException("Error storing the device " + d, e);
            } finally {
                DBTools.close(conn, stmt, null);
            }
            return true;
        }
        return false;
    }

    /**
     * Read the device from the data store.
     *
     * @param o The object that must be a device for this method
     *
     * @throws PersistentException in case of error reading the data store
     */
    public boolean read(Object o) throws PersistentStoreException {
        if (o instanceof Sync4jDevice) {
            Sync4jDevice d = (Sync4jDevice) o;

            Connection conn = null;
            PreparedStatement stmt = null;
            ResultSet rs = null;

            try {
                conn = dataSource.getConnection();

                stmt = conn.prepareStatement(sql[SQL_GET_DEVICE]);
                stmt.setString(1, d.getDeviceId());
                rs = stmt.executeQuery();

                if (rs.next() == false) {
                    throw new NotFoundException("Device not found for '"
                    + d.getDeviceId() + "'");
                }

                d.setDescription(rs.getString(1));
                d.setType       (rs.getString(2));
                String value = rs.getString(3);
                if (value == null) {
                    d.setClientNonce(null);
                } else if (value.equals("")) {
                    d.setClientNonce(new byte[0]);
                } else {
                    d.setClientNonce(Base64.decode(value.getBytes()));
                }
                value = rs.getString(4);
                if (value == null) {
                    d.setServerNonce(null);
                } else if (value.equals("")) {
                    d.setServerNonce(new byte[0]);
                } else {
                    d.setServerNonce(Base64.decode(value.getBytes()));
                }
                d.setServerPassword(rs.getString(5));
                d.setTimeZone(rs.getString(6));
                d.setConvertDate("Y".equals(rs.getString(7)));
                d.setCharset(rs.getString(8));


                Capabilities caps = new Capabilities();
                caps.setId                    (rs.getString(9));
                caps.setSupportLargeObject    ("Y".equals(rs.getString(10)));
                caps.setSupportNumberOfChanges("Y".equals(rs.getString(11)));
                d.setCapabilities(caps);

            } catch (SQLException e) {
                throw new PersistentStoreException("Error reading the device " + d, e);
            } finally {
                DBTools.close(conn, stmt, rs);
            }
            return true;
        }
        return false;
    }

    public Object[] read(Class objClass) throws PersistentStoreException {
        return null;
    }

    /**
     * Read all informations
     * @param o whichever object Sync
     * @param clause condition where for select
     *
     * @return Object[] array of object
     */
    public Object[] read(Object o, Clause clause) throws PersistentStoreException {
        if (o instanceof Sync4jDevice) {
            Connection conn = null;
            PreparedStatement stmt = null;
            ResultSet rs = null;

            ArrayList ret = new ArrayList();

            try {
                conn = dataSource.getConnection();

                PreparedWhere where = clause.getPreparedWhere();
                String query = sql[SQL_SELECT_ALL_DEVICES];
                if (where.sql.length() > 0) {
                    query += " where " + where.sql;
                }
                stmt = conn.prepareStatement(query);
                for (int i=0; i<where.parameters.length; ++i) {
                    stmt.setObject(i+1, where.parameters[i]);
                }
                rs = stmt.executeQuery();

                while (rs.next()) {
                    Sync4jDevice d = new Sync4jDevice();
                    d.setDeviceId      (rs.getString(1));
                    d.setDescription   (rs.getString(2));
                    d.setType          (rs.getString(3));

                    String value = rs.getString(4);
                    if (value == null) {
                        d.setClientNonce(null);
                    } else if (value.equals("")) {
                        d.setClientNonce(new byte[0]);
                    } else {
                        d.setClientNonce(Base64.decode(value.getBytes()));
                    }

                    value = rs.getString(5);
                    if (value == null) {
                        d.setServerNonce(null);
                    } else if (value.equals("")) {
                        d.setServerNonce(new byte[0]);
                    } else {
                        d.setServerNonce(Base64.decode(value.getBytes()));
                    }
                    d.setServerPassword(rs.getString(6));
                    d.setTimeZone(rs.getString(7));
                    d.setConvertDate("Y".equals(rs.getString(8)));
                    d.setCharset(rs.getString(9));

                    ret.add(d);
                }
                return ret.toArray(new Sync4jDevice[ret.size()]);
            } catch (SQLException e) {
                throw new PersistentStoreException("Error reading devices", e);
            } finally {
                DBTools.close(conn, stmt, rs);
            }
        }
        return null;
    }


    /**
     * Select the number of devices that satisfy the conditions specified in input.
     * @param clause The additional where clause.
     * @return int number of devices
     */
    public int count(Object o, Clause clause) throws PersistentStoreException {
        if (o instanceof Sync4jDevice) {
            Connection conn = null;
            PreparedStatement stmt = null;
            ResultSet rs = null;
            int n = 0;

            try {
                conn = dataSource.getConnection();

                PreparedWhere where = clause.getPreparedWhere();
                String query = sql[SQL_COUNT_DEVICES];
                if (where.sql.length()>0) {
                    query += " where " + where.sql;
                }
                stmt = conn.prepareStatement(query);
                for (int i=0; i<where.parameters.length; ++i) {
                    stmt.setObject(i+1, where.parameters[i]);
                }
                rs = stmt.executeQuery();

                while (rs.next()) {
                    n = rs.getInt(1);
                }
                return n;
            } catch (SQLException e) {
                throw new PersistentStoreException("Error reading count devices ", e);
            } finally {
                DBTools.close(conn, stmt, rs);
            }
        }
        return -1;
    }
    //---------------------------------------------------------- Private Methods

    /**
     * Return the next device id reading it by IdSpaceGenerator
     * @return int
     */
    private int getDeviceId() {
        IdSpaceGenerator idGenerator = new IdSpaceGenerator(NS_DEVICE);
        return Integer.parseInt(idGenerator.next());
    }
}

