/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package sync4j.server.store;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

import java.sql.*;

import sync4j.framework.tools.DBTools;

import sync4j.framework.server.LastTimestamp;
import sync4j.framework.server.store.Clause;
import sync4j.framework.server.store.PersistentStore;
import sync4j.framework.server.store.BasePersistentStore;
import sync4j.framework.server.store.NotFoundException;
import sync4j.framework.server.store.PersistentStoreException;

/**
 * This is the store for persistent information regarding the last time
 * stamp of the synchronization process.
 * It persists the following classes:
 * <ul>
 * <li>sync4j.framework.server.LastTimestamp</li>
 * </ul>
 *
 * @author Stefano Fornari @ Funambol
 * @author Harrie Hazewinkel
 * @version $Id: LastTimestampPersistentStore.java,v 1.4 2005/04/13 14:44:39 stefano_fornari Exp $
 */
public class LastTimestampPersistentStore
extends    BasePersistentStore
implements PersistentStore, java.io.Serializable {

    // --------------------------------------------------------------- Constants

    // -------------------------------------------------------------- Properties
    // --------------------------------------------- configurable SQL properties

    /**
     * The SQL query to insert the last timestamp
     */
    private String sqlInsertLastTimestamp =
    "insert into sync4j_last_sync (principal, sync_source, last_anchor_server, last_anchor_client, start_sync, end_sync) values(?, ?, ?, ?, ?, ?)";

    /** Getter for property sqlInsertLastTimestamp.
     * @return Value of property sqlInsertLastTimestamp.
     *
     */
    public String getSqlInsertLastTimestamp() {
        return sqlInsertLastTimestamp;
    }

    /** Setter for property sqlInsertLastTimestamp.
     * @param sqlInsertLastTimestamp New value of property sqlInsertLastTimestamp.
     *
     */
    public void setSqlInsertLastTimestamp(String sqlInsertLastTimestamp) {
        this.sqlInsertLastTimestamp = sqlInsertLastTimestamp;
    }

    /**
     * The SQL query to update the last timestamp
     */
    private String sqlUpdateLastTimestamp =
    "update sync4j_last_sync set last_anchor_server=?,last_anchor_client=?,start_sync=?,end_sync=? where principal=? and sync_source=?";

    /** Getter for property sqlUpdateLastTimestamp.
     * @return Value of property sqlUpdateLastTimestamp.
     *
     */
    public String getSqlUpdateLastTimestamp() {
        return sqlUpdateLastTimestamp;
    }

    /** Setter for property sqlUpdateLastTimestamp.
     * @param sqlUpdateLastTimestamp New value of property sqlUpdateLastTimestamp.
     *
     */
    public void setSqlUpdateLastTimestamp(String sqlUpdateLastTimestamp) {
        this.sqlUpdateLastTimestamp = sqlUpdateLastTimestamp;
    }

    /**
     * The SQL query to select the last timestamp
     */
    private String sqlSelectLastTimestamp =
    "select last_anchor_server, last_anchor_client,start_sync,end_sync from sync4j_last_sync where principal=? and sync_source=?";

    /** Getter for property sqlUpdateLastTimestamp.
     * @return Value of property sqlUpdateLastTimestamp.
     *
     */
    public String getSqlSelectLastTimestamp() {
        return sqlSelectLastTimestamp;
    }

    /** Setter for property sqlUpdateLastTimestamp.
     * @param sqlSelectLastTimestamp New value of property sqlUpdateLastTimestamp.
     *
     */
    public void setSqlSelectLastTimestamp(String sqlSelectLastTimestamp) {
        this.sqlSelectLastTimestamp = sqlSelectLastTimestamp;
    }

    // ------------------------------------------------------------ Private data

    // ------------------------------------------------------------ Constructors

    // ---------------------------------------------------------- Public methods

    /**
     * Store the last timestamp. 
     * Into tagServer there is the Next anchor sent by client: now this become
     * a Last anchor of the server.
     * Into tagClient there is the Next anchor sent by server: now this become
     * a Last anchor of the client.
     *
     * @param o the Object that must be a LastTimestamp object with all
     * information about start and end sync timestamp
     *
     * @throws PersistentStoreException
     */
    public boolean store(Object o)
    throws PersistentStoreException {
        if (o instanceof LastTimestamp) {
            LastTimestamp l = (LastTimestamp) o;

            Connection conn = null;
            PreparedStatement stmt = null;

            try {
                conn = dataSource.getConnection();

                stmt = conn.prepareStatement(sqlUpdateLastTimestamp);
                stmt.setString   (1, l.tagServer                );            
                stmt.setString   (2, l.tagClient                );
                stmt.setTimestamp(3, new Timestamp(l.start)     );
                stmt.setTimestamp(4, new Timestamp(l.end)       );
                stmt.setLong     (5, Long.parseLong(l.principal));
                stmt.setString   (6, l.database                 );
                int n = stmt.executeUpdate();

                if (n == 0) {
                    //
                    // The first time!!!
                    //
                    stmt.close();
                    stmt = conn.prepareStatement(sqlInsertLastTimestamp);
                    stmt.setLong     (1, Long.parseLong(l.principal));
                    stmt.setString   (2, l.database                 );
                    stmt.setString   (3, l.tagServer                );                
                    stmt.setString   (4, l.tagClient                );
                    stmt.setTimestamp(5, new Timestamp(l.start)     );
                    stmt.setTimestamp(6, new Timestamp(l.end)       );
                    stmt.executeUpdate();
                }
            } catch (SQLException e) {
                throw new PersistentStoreException("Error storing last timestamp", e);
            } finally {
                DBTools.close(conn, stmt, null);
            }
            return true;
        }
        return false;
    }

    public boolean read(Object o)
    throws PersistentStoreException {
        if (o instanceof LastTimestamp) {
            LastTimestamp l = (LastTimestamp) o;

            Connection conn = null;
            PreparedStatement stmt = null;
            ResultSet rs = null;

            try {
                conn = dataSource.getConnection();

                stmt = conn.prepareStatement(sqlSelectLastTimestamp);
                stmt.setLong(1, Long.parseLong(l.principal));
                stmt.setString(2, l.database);
                rs = stmt.executeQuery();

                if (rs.next() == false) {
                    throw new NotFoundException("Last timestamp not found for "
                    + l.toString()
                    );
                }
                l.tagServer = rs.getString   (1)          ;
                l.tagClient = rs.getString   (2)          ;
                l.start     = rs.getTimestamp(3).getTime();
                l.end       = rs.getTimestamp(4).getTime();
            } catch (SQLException e) {
                throw new PersistentStoreException("Error reading last timestamp", e);
            } finally {
                DBTools.close(conn, stmt, rs);
            }
            return true;
        }
        return false;
    }

    /** Read all objects stored the persistent media.
     *
     * @param objClass the object class handled by the persistent store
     *
     * @return an array containing the objects read. If no objects are found an
     *         empty array is returned. If the persistent store has not
     *         processed the quest, null is returned.
     *
     * @throws PersistentStoreException
     *
     */
    public Object[] read(Class objClass) throws PersistentStoreException {
        //
        // TO DO (not used yet)
        //
        return null;
    }


    public boolean delete(Object o) throws PersistentStoreException
    {
        return false;
    }

    public Object[] read(Object o, Clause clause) throws PersistentStoreException
    {
        return null;
    }

    public int count(Object o, Clause clause) throws PersistentStoreException
    {
        return -1;
    }

    public boolean store(Object o, String operation) throws PersistentStoreException
    {
        return false;
    }

}
