/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.transport.http.server;

import java.io.File;
import java.net.URL;
import java.util.Map;
import java.util.logging.Logger;
import java.util.logging.Level;

import sync4j.framework.transport.http.SyncHolder;
import sync4j.framework.server.SyncResponse;
import sync4j.framework.core.Sync4jException;
import sync4j.framework.config.ConfigClassLoader;
import sync4j.framework.config.ConfigurationException;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.logging.Sync4jLoggerName;
import sync4j.framework.server.error.ServerException;
import sync4j.framework.server.error.ServerFailureException;
import sync4j.framework.server.error.NotImplementedException;
import sync4j.framework.protocol.ProtocolException;

import sync4j.server.config.Configuration;
import sync4j.server.engine.SyncAdapter;

/**
 * Implements a <i>SyncHolder</i> that instance a SyncAdapter to handle a
 * synchronization request.
 *
 * @author  Luigia Fassina @ Funambol
 * @version $Id: LocalSyncHolder.java,v 1.11 2005/03/02 20:57:40 harrie Exp $
 */
public class LocalSyncHolder implements SyncHolder {
    
    // --------------------------------------------------------------- Constants
    
    // ------------------------------------------------------------ Private data
    
    private transient Logger log = Sync4jLogger.getLogger(Sync4jLoggerName.HANDLER);
    
    private SyncAdapter syncAdapter       = null;
    private long        creationTimestamp       ;
    private String      sessionId               ;
    
    // ------------------------------------------------------------ Constructors
    public LocalSyncHolder() throws ServerException {
        try {
            syncAdapter = new SyncAdapter(loadConfiguration());
        } catch (ConfigurationException e) {
            
        }
        creationTimestamp = System.currentTimeMillis();
    }
    
    // ---------------------------------------------------------- Public methods
    
   /** Processes an incoming XML message.
     *
     * @param requestData the SyncML request as an array of bytes
     * @param parameters SyncML request parameters
     * @param headers SyncML request headers
     *
     * @return the SyncML response as a <i>ISyncResponse</i> object
     *
     * @throws ServerException in case of a server error
     *
     */
    public SyncResponse processXMLMessage(final byte[] requestData, 
                                          final Map    parameters , 
                                          final Map    headers    ) 
    throws ServerException {
        return syncAdapter.processXMLMessage(requestData, parameters, headers);
    }
    
    /** Processes an incoming WBXML message.
     *
     * @param requestData the SyncML request as an array of bytes
     * @param parameters SyncML request parameters
     * @param headers SyncML request headers
     *
     * @return the SyncML response as a <i>ISyncResponse</i> object
     *
     * @throws ServerException in case of a server error
     *
     */
    public SyncResponse processWBXMLMessage(final byte[] requestData, 
                                            final Map    parameters , 
                                            final Map    headers    )
                                           
    throws ServerException {
        return syncAdapter.processWBXMLMessage(requestData, parameters, headers);
    }
    
    public void setSessionId(String sessionId) throws Sync4jException {
        this.sessionId = sessionId;
        syncAdapter.beginSync(sessionId);
    }
    
    public String getSessionId() {
        return this.sessionId;
    }
    
    /** Called when the SyncHolder is not required any more. It gives the holder
     * an opportunity to release and clean up resources.
     *
     * @throws java.lang.Exception in case of error. The real exception is stored
     * in the cause.
     *
     */
    public void close() throws Exception {
        syncAdapter.endSync();
    }
    
    /**
     * Returns the creation timestamp (in milliseconds since midnight, January
     * 1, 1970 UTC).
     *
     * @see sync4j.framework.transport.http.SyncHolder
     */
    public long getCreationTimestamp() {
        return creationTimestamp;
    }
    
    /**
     * Loads the configuration for the server.
     *
     * @throws ConfigurationException in case of errors.
     */
    private Configuration loadConfiguration()
    throws ConfigurationException {

        Configuration config = Configuration.getConfiguration();
        
        //
        // If config is null, being Configuration a singleton, only an exception
        // could have been happend.
        //
        if (config == null) {
            throw new ConfigurationException("Error loading configuration (see the log for details)!");
        }
        
        return config;
    }
}
